"""Module to collect exceptions we might throw.
"""

class UnableToFindUniqueTopic(Exception):

    def __init__(self, topic, count, extra):
        msg = 'Unable to find unique thread for topic %s. Found %i items. %s'%(
            topic, count, extra)
        Exception.__init__(self, msg)



"""Module for working comments from redis backend.
"""

import datetime
import doctest
import json
import logging
import redis

from eyap.core import comments


class RedisCommentThread(comments.CommentThread):
    """Class to represent a thread of redis comments.
    """

    def __init__(self, *args, ltrim=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.ltrim = ltrim
        self.redis = redis.Redis()

    def lookup_thread_id(self):
        return 'eyap:%s:%s' % (self.realm, self.topic)

    def add_comment(self, body, allow_create=False, allow_hashes=False,
                    summary=None):
        """Add comment as required by comments.CommentThread parent class.
        """
        thread_id = self.lookup_thread_id()
        if not allow_create and not self.redis.exists(thread_id):
            raise ValueError('Tried to add comment to non-exist thread %s' % (
                thread_id))

        comment = comments.SingleComment(
            self.user, datetime.datetime.now(datetime.timezone.utc), body, 
            summary=summary)
        lpush = self.redis.lpush(thread_id, comment.to_json())
        logging.debug('Pushing comment to redis returned %s', str(lpush))
        if self.ltrim:
            ltrim = self.redis.ltrim(thread_id, 0, self.ltrim)
            logging.debug('Redis ltrim returend %s', str(ltrim))
        else:
            ltrim = None

        return {'status': 'OK', 'lpush': lpush, 'ltrim': ltrim}

    def create_thread(self, body):
        return self.add_comment(body, allow_create=True)

    def delete_thread(self, really=False):
        thread_id = self.lookup_thread_id()        
        if not really:
            raise ValueError(
                'Cowardly refusing to delete thread %s since really=%s' % (
                    thread_id, really))
        self.redis.delete(thread_id)

    def lookup_comments(self, reverse=False):
        thread_id = self.lookup_thread_id()
        data = self.redis.lrange(thread_id, 0, -1)
        clist = []
        for jitem in data:
            citem = json.loads(jitem.decode('utf8'))
            clist.append(comments.SingleComment(**citem))
        if reverse:
            clist = list(reversed(clist))

        return comments.CommentSection(clist)

    def __getstate__(self):
        "Do not pickle redis connection"
        return {k: v for k, v in self.__dict__.items() if k != 'redis'}

    def __setstate__(self, state):
        for key, val in state.items():
            setattr(self, key, val)
        self.redis = redis.Redis()
        

    @staticmethod
    def _regr_test_pickle():
        """
>>> import pickle
>>> from eyap.core import redis_comments
>>> rc = redis_comments.RedisCommentThread(
...     'test-owner', 'test-realm', 'test-topic', 'test-owner')
>>> x=pickle.dumps(rc)
>>> y=pickle.loads(x)
"""

    @staticmethod
    def _regr_test():
        """
>>> from eyap.core import redis_comments
>>> rc = redis_comments.RedisCommentThread(
...     'test-owner', 'test-realm', 'test-topic', 'test-owner')
>>> rc.delete_thread(really=True)
>>> status = rc.add_comment('test_comment', allow_create=True)
>>> print(', '.join(['%s: %s' % (k, status[k]) for k in sorted(status)]))
lpush: 1, ltrim: None, status: OK
>>> sec = rc.lookup_comments('test-topic')
>>> print(sec.show())  # doctest: +ELLIPSIS
========================================
Subject: test_comment ...
Timestamp: ...
----------
test_comment
        """
    

if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests')



"""Module for handling reading comments.
"""

import datetime
import csv
import os
import doctest
import re
import urllib.parse
import json

import dateutil.parser

from requests.models import Response


class SingleComment(object):  # pylint: disable=too-many-instance-attributes
    """Class to hold information about a single comment.

    SingleComment instances will usually be collected into a list of
    instances in a CommentSection representing a thread of discussion.
    """

    def __init__(self, user, timestamp, body, url=None, summary=None,
                 markup=None, summary_cls=None):
        """Initializer.

        :arg user:        String name for user creating/owning comment.

        :arg timestamp:   String timestamp for when comment last edited.

        :arg body:        String body of comment.

        :arg url=None:    String url where comment lives.

        :arg summary=None:    One line text summary of comment. If not given,
                              we take first 40 characters of 1st line of body.

        :arg markup=None:     Text of comment marked up with HTML or
                              other way of display. If not provided, we just
                              use body.

        :arg summary_cls=None:  Optional string representing CSS class for
                                summary. You can set this as late as you like.
                                If provided, comment_view.html will apply this
                                class to the summary which can be helpful in
                                making certain posts stand out.

        """
        self.user = user
        self.timestamp = timestamp
        self.body = body
        self.markup = markup if markup else body
        self.url = url
        self.summary = summary if summary else (
            body.split('\n')[0][0:40] + ' ...')
        self.summary_cls = None
        self.display_timestamp = timestamp

    def make_anchor_id(self):
        """Return string to use as URL anchor for this comment.
        """
        result = re.sub(
            '[^a-zA-Z0-9_]', '_', self.user + '_' + self.timestamp)
        return result

    def make_url(self, my_request, anchor_id=None):
        """Make URL to this comment.

        :arg my_request:  The request object where this comment is seen from.

        :arg anchor_id=None:    Optional anchor id. If None, we use
                                self.make_anchor_id()

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       String URL to this comment.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Be able to create links to this comment.

        """
        if anchor_id is None:
            anchor_id = self.make_anchor_id()
        result = '{}?{}#{}'.format(
            my_request.path, urllib.parse.urlencode(my_request.args),
            anchor_id)

        return result

    def to_dict(self):
        """Return description of self in dict format.

        This is useful for serializing to something like json later.
        """
        jdict = {
            'user': self.user,
            'summary': self.summary,
            'body': self.body,
            'markup': self.markup,
            'url': self.url,
            'timestamp': self.timestamp
            }
        return jdict

    def set_display_mode(self, mytz, fmt):
        """Set the display mode for self.

        :arg mytz:        A pytz.timezone object.

        :arg fmt:         A format string for strftime.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Modifies self.display_timestamp to first parse
                        self.timestamp and then format according to given
                        timezone and format string.

        """
        my_stamp = dateutil.parser.parse(self.timestamp)
        tz_stamp = my_stamp.astimezone(
            mytz) if my_stamp.tzinfo is not None else my_stamp
        self.display_timestamp = tz_stamp.strftime(fmt)

    def to_json(self):
        my_dict = self.to_dict()
        my_ts = my_dict['timestamp']
        my_dict['timestamp'] = getattr(my_ts, 'isoformat', lambda: my_ts)()
        return json.dumps(my_dict)

    def __str__(self):
        return 'Subject: %s\nTimestamp: %s\n%s\n%s' % (
            self.summary, self.display_timestamp, '-'*10, self.body)


class CommentSection(object):
    """Class to represent a section, thread or other collection of comments.

    A CommentSection has the following useful features:

       self.comments:  Sequence of SingleComment instances with the comments.
       self.display:   Function to dislay the comments.

    """

    def __init__(self, comments):
        """Initializer.

        :arg comments:        Sequence of SingleComment instances representing
                              a thread or other collection of related comments.

        """
        self.comments = comments

    def set_display_mode(self, mytz, fmt="%Y-%m-%d %H:%M:%S %Z%z"):
        """Call set_display_mode on items in self.comments with given args.

        :arg mytz:      A pytz.timezone object.

        :arg fmt="%Y-%m-%d%H:%M:%S%Z%z":   Optional format string.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Call set_display_mode(mytz, fmt) for everything in
                        self.comments.

        """
        for item in self.comments:
            item.set_display_mode(mytz, fmt)

    def show(self):
        "Return string showing the comments."
        result = '\n'.join(['%s\n%s' % ('=' * 40, c) for c in self.comments])
        return result

    def __str__(self):
        return self.show()


class CommentThread(object):
    """Abstract class used to interact with discussion threads.

    The idea is that we have the CommentThread as an abstract class which
    various backends can implement. Then you can use say GitHub issues as
    your comment thread or switch easily to a local database or cloud
    database or whatever relatively painlessly.
    """

    # Regular expression for what consistutes a hashtag.
    hashtag_re = r'\s+(#[-.a-zA-Z_/]+[a-zA-Z_])'

    # Regular expression for valid attachment location
    valid_attachment_loc_re = '^[-.a-zA-Z0-9_,/]+$'

    def __init__(self, owner, realm, topic,
                 user=None, token=None, thread_id=None):
        """Initializer.

        Since this is an abstract class which each backend implements,
        it is not quite possible to know all the different arguments which
        may be necessary. We use the following generic arguments that all
        sub-class of CommentThread are expected to implement. Your particular
        sub-class may also add additional arguments if necessary.

        :arg owner:    String owner (e.g., the repository owner if using
                       GitHub as a backend).

        :arg realm:    The realm (e.g., repository name on GitHub).

        :arg topic:    Topic or title of comment thread.

        :arg user:     String user name.

        :arg token:    Password or token to use to connect to backend.

        :arg thread_id=None: Optional thread id. Can leave as None if you want
                             this dynamically looked up via lookup_thread_id.

        """
        self.owner = owner
        self.realm = realm
        self.topic = topic
        self.user = user
        self.token = token
        self.thread_id = thread_id
        self.content = None

    def lookup_thread_id(self):
        """Find the thread id for the given owner, realm, topic, etc.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       The thread id (either integer, string, etc., depending
                        on the backend) used for the given thread indicated
                        by args to init.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Basically this serves to lookup the thread id if/when
                        necessary so we can interact with the comments.
                        It is an abstract method so that we can write higher
                        level methods in the CommentThread object which need
                        it.

        """
        raise NotImplementedError

    def add_comment(self, body, allow_create=False, allow_hashes=False,
                    summary=None):
        """Add the string comments to the thread.

        :arg body:        String/text of comment to add.

        :arg allow_create=False: Whether to automatically create a new thread
                                 if a thread does not exist (usually by calling
                                 self.create_thread).

        :arg allow_hashes=False: Whether to support hashtag mentions of other
                                 topics and automatically insert comment in
                                 body into those topics as well.

                                 *IMPORTANT*: if you recursively call
                                 add_comment to insert the hashes, you should
                                 make sure to set this to False to prevent
                                 infinite hash processing loops.

        :arg summary=None:       Optional summary to use for the comment. The
                                 backend should pass this to SingleComment if
                                 given.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       Usually a response object from an HTTP call indicating
                        what happend. Otherwise, somethng with a status_code
                        and reason argument will suffice.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Add a comment to the discussion thread.

        """
        raise NotImplementedError

    def lookup_comments(self, reverse=False):
        """Return CommentSection instance showing all comments for thread.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Require backend implementations to do this so it can
                        be called in higher level functions like
                        get_comment_section.

                        Note that the user probably should call
                        get_comment_section and not call this directly.
        """
        raise NotImplementedError

    def create_thread(self, body):
        """Create a new thread.

        :arg body:        Text for the initial comment in the thread.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       A response object indicating what happened.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Create a thread for self.topic.

        """
        raise NotImplementedError

    def get_comment_section(self, force_reload=False, reverse=False):
        """Get CommentSection instance representing all comments for thread.

        :arg force_reload=False:  Whether to force reloading comments
                                  directly or allow using what is cached
                                  in self.content if possible.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       CommentSection representing all comments for thread.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        High-level function called by user to get comments.

        """
        if self.content is not None and not force_reload:
            return self.content
        if self.thread_id is None:
            self.thread_id = self.lookup_thread_id()
        self.content = self.lookup_comments(reverse=reverse)

        return self.content

    def validate_attachment_location(self, location):
        """Validate a proposed attachment location.

        :arg location:     String representing location to put attachment.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Raises an exception if attachment location is bad.
                   By default, this just forces reasonable characters.
                   Sub-classes can override as desired.

        """
        if not re.compile(self.valid_attachment_loc_re).match(location):
            raise ValueError(
                'Bad chars in attachment location. Must match %s' % (
                    self.valid_attachment_loc_re))

    def upload_attachment(self, location, data):
        """Upload an attachment.

        :arg location:    String identifying name or location to store
                          the attachment. Usually just a file name is
                          a good thing to use.

        :arg data:        Either str or bytes or a file-like object with
                          a read method represneting data for the attachment.
                          If a string, it must be a base 64 encoded
                          representation of the bytes. If bytes, then we
                          will base64.b64encode(data).decode('ascii').

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:    A string that can be added to a comment to reference
                     the attachment.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:     Provide a way to upload a file attachment.  Ideally
                     sub-classes should call validate_attachment_location
                     to make sure the proposed location is valid.

        """
        raise NotImplementedError


class FileCommentThread(CommentThread):
    """Example file based comment thread for testing and showing example usage.

    This is a subclass of CommentThread which implements the required functions
    to store the comments in files. The realm argument to __init__ must be
    an existing directory. The topic argument to __init__ will be the name
    of the file to store comments in.
    """

    header = ['user', 'timestamp', 'summary', 'body', 'url']  # header for csv

    def lookup_thread_id(self):
        "Lookup the thread id as path to comment file."

        path = os.path.join(self.realm, self.topic + '.csv')
        return path

    def lookup_comments(self, reverse=False):
        "Implement as required by parent to lookup comments in file system."

        comments = []
        if self.thread_id is None:
            self.thread_id = self.lookup_thread_id()
        path = self.thread_id
        with open(self.thread_id, 'r', newline='') as fdesc:
            reader = csv.reader(fdesc)
            header = reader.__next__()
            assert header == self.header
            for num, line in enumerate(reader):
                if not line:
                    continue
                assert len(line) == len(header), (
                    'Line %i in path %s misformatted' % (num+1, path))
                line_kw = dict(zip(header, line))
                comments.append(SingleComment(**line_kw))
        if reverse:
            comments = list(reversed(comments))

        return CommentSection(comments)

    def create_thread(self, body):
        """Implement create_thread as required by parent.

        This basically just calls add_comment with allow_create=True
        and then builds a response object to indicate everything is fine.
        """
        self.add_comment(body, allow_create=True)
        the_response = Response()
        the_response.code = "OK"
        the_response.status_code = 200

    def add_comment(self, body, allow_create=False, allow_hashes=False,
                    summary=None):
        "Implement as required by parent to store comment in CSV file."

        if allow_hashes:
            raise ValueError('allow_hashes not implemented for %s yet' % (
                self.__class__.__name__))
        if self.thread_id is None:
            self.thread_id = self.lookup_thread_id()
        if not os.path.exists(self.thread_id):
            if not allow_create:
                raise KeyError(self.topic)
            with open(self.thread_id, 'a', newline='') as fdesc:
                csv.writer(fdesc).writerow(self.header)

        with open(self.thread_id, 'a', newline='') as fdesc:
            writer = csv.writer(fdesc)
            writer.writerow([self.user, datetime.datetime.utcnow(), summary,
                             body, ''])

    @staticmethod
    def _regr_tests():
        """
>>> import os, tempfile, shutil
>>> from eyap.core import comments
>>> tempdir = tempfile.mkdtemp()
>>> owner, realm, topic, user = 'test_owner', tempdir, 'test_topic', 't_user'
>>> ctest = comments.FileCommentThread(owner, realm, topic, user, None)
>>> ctest.add_comment('testing comment%ccool stuff here' % 10, allow_create=1)
>>> print(ctest.lookup_comments()) # doctest: +ELLIPSIS
========================================
Subject: testing comment ...
Timestamp: ...
----------
testing comment
cool stuff here
>>> shutil.rmtree(tempdir)
>>> os.path.exists(tempdir)
False
"""

if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests.')



"""Module for working comments from GitHub backend.
"""

import datetime
import doctest
import time
import collections
import re
import json
import logging
import zipfile
import base64

import requests

from eyap.core import comments, yap_exceptions


def fake_markdown(text, *args, **kw):
    "Fake markdown"
    _dummy_ignore = args, kw
    return text


try:
    from markdown import markdown
except ImportError as problem:
    logging.warning('\n'.join([
        'Could not import markdown package. Will render as plain.',
        'Install the markdown package if you want comments rendered as',
        'markdown.']))

    def markdown(text, *args, **kw):
        "Fake markdown by just return input text"
        dummy = args, kw
        return text


GitHubInfo = collections.namedtuple('GitHubInfo', [
    'owner', 'realm', 'user', 'token'])


class GitHubAngry(Exception):
    """Exception to indicate something wrong with github API.
    """

    def __init__(self, msg, *args, **kw):
        Exception.__init__(self, msg, *args, **kw)


class GitHubCommentGroup(object):
    """Class to represent a group of github comments.
    """

    def __init__(self, topic_re, gh_info, max_threads=None, params=None):
        """Initializer.

        :arg gh_info:    Instance of GitHubInfo describing how to access github
        :arg param=None: Optional dict of params to pass to github request

        """
        self.topic_re = topic_re
        self.gh_info = gh_info
        self.max_threads = max_threads
        self.base_url = 'https://api.github.com/repos/%s/%s' % (
            self.gh_info.owner, self.gh_info.realm)
        self.params = dict(params) if params else {}

    @staticmethod
    def parse_date(my_date):
        """Parse a date into canonical format of datetime.dateime.

        :param my_date:    Either datetime.datetime or string in 
                           '%Y-%m-%dT%H:%M:%SZ' format.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A datetime.datetime.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Parse a date and make sure it has no time zone.

        """
        if isinstance(my_date, datetime.datetime):
            result = my_date
        elif isinstance(my_date, str):
            result = datetime.datetime.strptime(my_date, '%Y-%m-%dT%H:%M:%SZ')
        else:
            raise ValueError('Unexpected date format for "%s" of type "%s"' % (
                str(my_date), type(my_date)))
        assert result.tzinfo is None, 'Unexpected tzinfo for date %s' % (
            result)
        return result

    def get_thread_info(self, enforce_re=True, latest_date=None):
        """Return a json list with information about threads in the group.

        :param enforce_re=True:        Whether to require titles to match
                                       regexp in self.topic_re.

        :param latest_date=None:       Optional datetime.datetime for latest
                                       date to consider. Things past this
                                       are ignored.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  List of github items found.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Return a json list with information about threads
                   in the group. Along with latest_date, this can be used
                   to show issues.

        """
        result = []
        my_re = re.compile(self.topic_re)
        url = '%s/issues?sort=updated' % (self.base_url)
        latest_date = self.parse_date(latest_date) if latest_date else None
        while url:
            kwargs = {} if not self.gh_info.user else {'auth': (
                self.gh_info.user, self.gh_info.token)}
            my_req = requests.get(url, params=self.params, **kwargs)
            my_json = my_req.json()
            for item in my_json:
                if (not enforce_re) or my_re.search(item['title']):
                    idate = self.parse_date(item['updated_at'])
                    if (latest_date is not None and idate > latest_date):
                        logging.debug('Skip %s since updated at %s > %s',
                                      item['title'], idate, latest_date)
                        continue
                    result.append(item)
                    if self.max_threads is not None and len(
                            result) >= self.max_threads:
                        logging.debug('Stopping after max_threads=%i threads.',
                                      len(result))
                        return result
            url = None
            if 'link' in my_req.headers:
                link = my_req.headers['link'].split(',')
                for thing in link:
                    potential_url, part = thing.split('; ')
                    if part == 'rel="next"':
                        url = potential_url.lstrip(' <').rstrip('> ')

        return result

    def export(self, out_filename):
        """Export desired threads as a zipfile to out_filename.
        """
        with zipfile.ZipFile(out_filename, 'w', zipfile.ZIP_DEFLATED) as arc:
            id_list = list(self.get_thread_info())
            for num, my_info in enumerate(id_list):
                logging.info('Working on item %i : %s', num, my_info['number'])
                my_thread = GitHubCommentThread(
                    self.gh_info.owner, self.gh_info.realm, my_info['title'],
                    self.gh_info.user, self.gh_info.token,
                    thread_id=my_info['number'])
                csec = my_thread.get_comment_section()
                cdict = [item.to_dict() for item in csec.comments]
                my_json = json.dumps(cdict)
                arc.writestr('%i__%s' % (my_info['number'], my_info['title']),
                             my_json)

    @staticmethod
    def _test_export():
        """Simple regression test to make sure export works.

    NOTE: this test will hit the github web site unauthenticated. There are
          pretty tight rate limits for that so if you are re-running this
          test repeatedly, it will fail. To manually verify you can set
          user and token and re-run.

>>> user, token = None, None
>>> import tempfile, shlex, os, zipfile
>>> from eyap.core import github_comments
>>> info = github_comments.GitHubInfo('octocat', 'Hello-World', user, token)
>>> group = github_comments.GitHubCommentGroup('.', info, max_threads=3)
>>> fn = tempfile.mktemp(suffix='.zip')
>>> group.export(fn)
>>> zdata = zipfile.ZipFile(fn)
>>> len(zdata.filelist)
3
>>> data = zdata.read(zdata.infolist()[0].filename)
>>> len(data) > 10
True
>>> del zdata
>>> os.remove(fn)
>>> os.path.exists(fn)
False
"""


class GitHubCommentThread(comments.CommentThread):
    """Sub-class of CommentThread using GitHub as a back-end.
    """

    __thread_id_cache = {}

    # Base url to use in searching for issues.
    search_url = 'https://api.github.com/search/issues'
    url_extras = ''  # useful in testing to add things to URL

    def __init__(self, *args, attachment_location='files', **kw):
        """Initializer.

        :arg *args, **kw:  As for CommentThread.__init__.

        """
        comments.CommentThread.__init__(self, *args, **kw)
        self.base_url = 'https://api.github.com/repos/%s/%s' % (
            self.owner, self.realm)
        self.attachment_location = attachment_location

    @classmethod
    def sleep_if_necessary(cls, user, token, endpoint='search', msg=''):
        """Sleep a little if hit github recently to honor rate limit.
        """
        my_kw = {'auth': (user, token)} if user else {}
        info = requests.get('https://api.github.com/rate_limit', **my_kw)
        info_dict = info.json()
        remaining = info_dict['resources'][endpoint]['remaining']
        logging.debug('Search remaining on github is at %s', remaining)

        if remaining <= 5:
            sleep_time = 120
        else:
            sleep_time = 0
        if sleep_time:
            logging.warning('Sleep %i since github requests remaining  = %i%s',
                            sleep_time, remaining, msg)
            time.sleep(sleep_time)
            return True

        return False

    @classmethod
    def update_cache_key(cls, cache_key, item=None):
        """Get item in cache for cache_key and add item if item is not None.
        """
        contents = cls.__thread_id_cache.get(cache_key, None)
        if item is not None:
            cls.__thread_id_cache[cache_key] = item

        return contents

    @classmethod
    def lookup_cache_key(cls, cache_key):
        "Syntactic sugar for update_cache_key(cache_key)"

        return cls.update_cache_key(cache_key)

    def lookup_thread_id(self):
        """Lookup thread id as required by CommentThread.lookup_thread_id.

        This implementation will query GitHub with the required parameters
        to try and find the topic for the owner, realm, topic, etc., specified
        in init.
        """

        query_string = 'in:title "%s" repo:%s/%s' % (
            self.topic, self.owner, self.realm)
        cache_key = (self.owner, self.realm, self.topic)
        result = self.lookup_cache_key(cache_key)
        if result is not None:
            my_req = self.raw_pull(result)
            if my_req.status_code != 200:
                result = None  # Cached item was no good
            elif my_req.json()['title'] != self.topic:
                logging.debug('Title must have changed; ignore cache')
                result = None
            else:
                logging.debug('Using cached thread id %s for %s', str(result),
                              str(cache_key))
                return result
        data, dummy_hdr = self.raw_search(self.user, self.token, query_string)

        if data['total_count'] == 1:   # unique match
            if data['items'][0]['title'] == self.topic:
                result = data['items'][0]['number']
            else:
                result = None
        elif data['total_count'] > 1:  # multiple matches since github doesn't
            searched_data = [          # have unique search we must filter
                item for item in data['items'] if item['title'] == self.topic]
            if not searched_data:  # no matches
                return None
            elif len(searched_data) > 1:
                raise yap_exceptions.UnableToFindUniqueTopic(
                    self.topic, data['total_count'], '')
            else:
                assert len(searched_data) == 1, (
                    'Confused searching for topic "%s"' % str(self.topic))
                result = searched_data[0]['number']
        else:
            result = None
        self.update_cache_key(cache_key, result)

        return result

    @classmethod
    def raw_search(cls, user, token, query, page=0):
        """Do a raw search for github issues.

        :arg user:        Username to use in accessing github.

        :arg token:       Token to use in accessing github.

        :arg query:       String query to use in searching github.

        :arg page=0:      Number of pages to automatically paginate.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       The pair (result, header) representing the result
                        from github along with the header.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Search for issues on github. If page > 0 then we
                        will pull out up to page more pages via automatic
                        pagination. The best way to check if you got the
                        full results is to check if results['total_count']
                        matches len(results['items']).

        """
        page = int(page)
        kwargs = {} if not user else {'auth': (user, token)}
        my_url = cls.search_url
        data = {'items': []}
        while my_url:
            cls.sleep_if_necessary(
                user, token, msg='\nquery="%s"' % str(query))
            my_req = requests.get(my_url, params={'q': query}, **kwargs)
            if my_req.status_code != 200:
                raise GitHubAngry(
                    'Bad status code %s finding query %s because %s' % (
                        my_req.status_code, query, my_req.reason))
            my_json = my_req.json()
            assert isinstance(my_json['items'], list)
            data['items'].extend(my_json.pop('items'))
            data.update(my_json)
            my_url = None

            if page and my_req.links.get('next', False):
                my_url = my_req.links['next']['url']
            if my_url:
                page = page - 1
                logging.debug(
                    'Paginating %s in raw_search (%i more pages allowed)',
                    my_req.links, page)

        return data, my_req.headers

    def raw_pull(self, topic):
        """Do a raw pull of data for given topic down from github.

        :arg topic:    String topic (i.e., issue title).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:      Result of request data from github API.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:       Encapsulate call that gets raw data from github.

        """
        assert topic is not None, 'A topic of None is not allowed'
        kwargs = {} if not self.user else {'auth': (self.user, self.token)}
        my_req = requests.get('%s/issues/%s' % (
            self.base_url, topic), **kwargs)
        return my_req

    def lookup_comment_list(self):
        """Lookup list of comments for an issue.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:  The pair (ISSUE, COMMENTS) where ISSUE is a dict for the
                   main issue and COMMENTS is a list of comments on the issue.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Do the work of getting data from github, handling paging,
                   and so on.

        """
        if self.thread_id is None:
            return None, None

        # Just pulling a single issue here so pagination shouldn't be problem
        my_req = self.raw_pull(self.thread_id)
        if my_req.status_code != 200:
            raise GitHubAngry('Bad status code %s because %s' % (
                my_req.status_code, my_req.reason))
        issue_json = my_req.json()
        comments_url = issue_json['comments_url'] + self.url_extras
        kwargs = {} if not self.user else {'auth': (self.user, self.token)}
        comments_json = []
        while comments_url:
            logging.debug('Pulling comments URL: %s', comments_url)
            c_req = requests.get(comments_url, **kwargs)
            my_json = c_req.json()
            assert isinstance(my_json, list)
            comments_json.extend(my_json)
            comments_url = None
            if 'link' in c_req.headers:  # need to handle pagination.
                logging.debug('Paginating in lookup_comment_list')
                link = c_req.headers['link'].split(',')
                for thing in link:
                    potential_url, part = thing.split('; ')
                    if part == 'rel="next"':
                        comments_url = potential_url.lstrip(' <').rstrip('> ')

        return issue_json, comments_json

    def lookup_comments(self, reverse=False):
        if self.thread_id is None:
            self.thread_id = self.lookup_thread_id()
        issue_json, comment_json = self.lookup_comment_list()
        if issue_json is None and comment_json is None:
            return comments.CommentSection([])
        cthread_list = [comments.SingleComment(
            issue_json['user']['login'], issue_json['created_at'],
            issue_json['body'], issue_json['html_url'],
            markup=markdown(issue_json['body'], extensions=[
                'fenced_code', 'tables', 'markdown.extensions.nl2br']))]

        for item in comment_json:
            comment = comments.SingleComment(
                item['user']['login'], item['updated_at'], item['body'],
                item['html_url'], markup=markdown(
                    item['body'], extensions=[
                        'fenced_code', 'tables', 'markdown.extensions.nl2br']))

            cthread_list.append(comment)

        if reverse:
            cthread_list = list(reversed(cthread_list))

        return comments.CommentSection(cthread_list)

    def add_comment(self, body, allow_create=False, allow_hashes=True,
                    summary=None):
        """Implement as required by CommentThread.add_comment.

        :arg body:    String/text of comment to add.

        :arg allow_create=False: Whether to automatically create a new thread
                                 if a thread does not exist (usually by calling
                                 self.create_thread).

        :arg allow_hashes=True:  Whether to support hashtag mentions of other
                                 topics and automatically insert comment in
                                 body into those topics as well.

                                 *IMPORTANT*: if you recursively call
                                 add_comment to insert the hashes, you should
                                 make sure to set this to False to prevent
                                 infinite hash processing loops.

        arg summary=None:        Optional summary. If not given, we will
                                 extract one from body automatically if
                                 necessary.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:     Response object indicating whether added succesfully.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:      This uses the GitHub API to try to add the given comment
                      to the desired thread.

        """
        if self.thread_id is None:
            self.thread_id = self.lookup_thread_id()
        data = json.dumps({'body': body})
        if self.thread_id is None:
            if allow_create:
                return self.create_thread(body)
            else:
                raise ValueError(
                    'Cannot find comment existing comment for %s' % self.topic)

        result = requests.post('%s/issues/%s/comments' % (
            self.base_url, self.thread_id), data, auth=(self.user, self.token))
        if result.status_code != 201:
            if result.reason == 'Not Found' and allow_create:
                return self.create_thread(body)
            else:
                raise GitHubAngry(
                    'Bad status %s add_comment on %s because %s' % (
                        result.status_code, self.topic, result.reason))

        if allow_hashes:
            self.process_hashes(body)

        return result

    def process_hashes(self, body, allow_create=False):
        """Process any hashes mentioned and push them to related topics.

        :arg body:    Body of the comment to check for hashes and push out.

        :arg allow_create=False:  Whether to allow creating new topics
                                  from hash tag mentions.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:    Look for hashtags matching self.hashtag_re and when found,
                    add comment from body to those topics.

        """
        hash_re = re.compile(self.hashtag_re)
        hashes = hash_re.findall(body)
        done = {self.topic.lower(): True}
        for mention in hashes:
            mention = mention.strip('#')
            if mention.lower() in done:
                continue  # Do not duplicate hash mentions
            new_thread = self.__class__(
                owner=self.owner, realm=self.realm, topic=mention,
                user=self.user, token=self.token)
            my_comment = '# Hashtag copy from %s:\n%s' % (self.topic, body)
            new_thread.add_comment(
                my_comment, allow_create=allow_create,
                allow_hashes=False)  # allow_hashes=False to prevent inf loop
            done[mention.lower()] = True

    def create_thread(self, body):
        data = json.dumps({'body': body, 'title': self.topic})
        result = requests.post('%s/issues' % (self.base_url),
                               data, auth=(self.user, self.token))
        if result.status_code != 201:
            raise GitHubAngry(
                'Bad status %s in create_thread on %s because %s' % (
                    result.status_code, self.topic, result.reason))

        return result

    def upload_attachment(self, location, data):
        """Upload attachment as required by CommentThread class.

        See CommentThread.upload_attachment for details.
        """
        self.validate_attachment_location(location)
        content = data.read() if hasattr(data, 'read') else data
        orig_content = content
        if isinstance(content, bytes):
            content = base64.b64encode(orig_content).decode('ascii')
        else:
            pass  # Should be base64 encoded already
        apath = '%s/%s' % (self.attachment_location, location)
        url = '%s/contents/%s' % (self.base_url, apath)
        result = requests.put(
            url, auth=(self.user, self.token), data=json.dumps({
                'message': 'file attachment %s' % location,
                'content': content}))
        if result.status_code != 201:
            raise ValueError(
                "Can't upload attachment %s due to error %s." % (
                    location, result.reason))
        return '[%s](https://github.com/%s/%s/blob/master/%s)' % (
            location, self.owner, self.realm, apath)

    @staticmethod
    def _regr_test_lookup():
        """
    NOTE: this test will hit the github web site unauthenticated. There are
          pretty tight rate limits for that so if you are re-running this
          test repeatedly, it will fail. To manually verify you can set
          user and token and re-run.

>>> user, token = None, None
>>> import tempfile, shlex, os, zipfile
>>> from eyap.core import github_comments
>>> t = github_comments.GitHubCommentThread(
...     'emin63', 'eyap', user, token, thread_id='1')
>>> i, c = t.lookup_comment_list()
>>> t.url_extras = '?per_page=1'
>>> more_i, more_c = t.lookup_comment_list()
>>> i == more_i and c == more_c
True
>>> t.url_extras = ''
        """

if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests')



"""Extended yap package for comment management.

With eyap, you can choose a backend to manage your comments. A key goal of
the eyap package is that you can write your code in a generic way and easily
switch the back-end. For example, you can use a simple file back-end with
something like like the following:

```

>>> import eyap, tempfile, os, shutil  # Import some things used for demo.
>>> backend = 'file'                   #  Use a simple file backend for demo.
>>> finfo = {'owner': 'emin63', 'realm': tempfile.mkdtemp(), 'topic': 'test'}
>>> comment_thread = eyap.Make.comment_thread('file', **finfo)
>>> comment_thread.add_comment('Testing things', allow_create=True)
>>> comment_thread.add_comment('is great!')

```

The code above will create a new comment thread and add a comment to it
while also creating the thread if it does not exist. Later, you could
access the thread via something like:

```
>>> print(str(comment_thread.lookup_comments())) # doctest: +ELLIPSIS
========================================
Subject: Testing things ...
Timestamp: ...
----------
Testing things
========================================
Subject: is great! ...
Timestamp: ...
----------
is great!

```

The main reason for the existence of something like eyap is so that you
could use pretty much the same exact code with a different backend. For
example, using 'github' instead of 'file' and setting realm/owner to the
organization/repo in github while providing a valid user/token for github
access will post or read an issue from github.

```

>>> ginfo = {'owner': 'emin63', 'realm': 'eyap', 'topic': 'Describe usage'}
>>> g_thread = eyap.Make.comment_thread('github', **ginfo)
>>> print(str(g_thread.lookup_comments())) # doctest: +ELLIPSIS
========================================
Subject: We need a simple description of how to u ...
Timestamp: ...
----------
We need a simple description of how to use eyap.
========================================
Subject: Start with top-level README.md ...
Timestamp: ...
----------
Start with top-level README.md
========================================
Subject: All done! ...
Timestamp: ...
----------
All done!

```

Note that in the above, we only read from github since we did not provide
any username or password. If you had a username and token or password, you
could post comments to github as well.

   Plesae do *NOT* post to above issue; use your own repo for tests.  :)

Finally, we cleanup the file based backend since we don't need it anymore.

```

>>> shutil.rmtree(finfo['realm'])  # Remove the directory to cleanup test.
>>> os.path.exists(finfo['realm']) # Verify we cleaned up.
False

```

See the [README.md on
github](https://github.com/emin63/eyap/blob/master/README.md) for
further details on usage, how to create new backends, etc.

"""

VERSION = '0.8.4'

import logging
import doctest

from eyap.core import comments, github_comments

try:  # Try to import redis if possible
    from eyap.core import redis_comments
except ImportError as prob:  # Create mock class to complain about install.
    msg = '\n'.join(['Could not import redis_comments because of error:',
                     str(prob), 'If the problem is that redis is missing,'
                     'Consider installing via `pip install redis`.'])
    logging.debug(msg)
    class ComplainOnRedis:
        "Mock module to complain if we try to use redis backend."
        class RedisCommentThread:
            "Mock class to complain if we try to use redis backend."
            def __init__(self, *args, **kwargs):
                dummy = args, kwargs
                raise Exception(msg)

    redis_comments = ComplainOnRedis


class Make(object):

    _known_backends = {
        'file': comments.FileCommentThread,
        'github': github_comments.GitHubCommentThread,
        'redis': redis_comments.RedisCommentThread
        }

    @classmethod
    def comment_thread(cls, backend, *args, **kwargs):
        """Create a comment thread for the desired backend.

        :arg backend:        String name of backend (e.g., 'file',
                             'github', 'redis', etc.).

        :arg *args, **kwargs:   Arguments to be passed to contructor
                                for that backend.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:  A CommentThread sub-class for the given backend.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Some simple syntatic sugar for creating the desired
                  backend.

        """
        ct_cls = cls._known_backends.get(backend)
        if not ct_cls:
            return None
        return ct_cls(*args, **kwargs)

if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')
